import pytest
import torch

from vesin.torch import NeighborList


@pytest.mark.parametrize("full_list", [True, False])
@pytest.mark.parametrize("requires_grad", [(True, True), (False, True), (True, False)])
@pytest.mark.parametrize("quantities", ["ijS", "D", "d", "ijSDd"])
def test_autograd(full_list, requires_grad, quantities):
    torch.manual_seed(0xDEADBEEF)

    points_fractional = torch.rand((34, 3), dtype=torch.float64)
    box = torch.diag(5 * torch.rand(3, dtype=torch.float64))
    box += torch.rand((3, 3), dtype=torch.float64)

    points = points_fractional @ box

    points.requires_grad_(requires_grad[0])
    box.requires_grad_(requires_grad[1])

    calculator = NeighborList(cutoff=7.8, full_list=full_list)

    def compute(points, box):
        results = calculator.compute(points, box, periodic=True, quantities=quantities)
        return results

    torch.autograd.gradcheck(compute, (points, box), fast_mode=True)



import torch

from .. import NeighborList as NeighborListTorch


try:  # only define a metatensor adapter if metatensor is available
    from metatensor.torch import Labels, TensorBlock
    from metatensor.torch.atomistic import NeighborListOptions, System

    _HAS_METATENSOR = True
except ModuleNotFoundError:
    _HAS_METATENSOR = False

    class Labels:
        pass

    class TensorBlock:
        pass

    class System:
        pass

    class NeighborListOptions:
        pass


class NeighborList:
    """
    A neighbor list calculator that can be used with metatensor's atomistic models.

    The main difference with the other calculators is the automatic handling of
    different length unit between what the model expects and what the ``System`` are
    using.

    .. seealso::

        The :py:func:`vesin.torch.metatensor.compute_requested_neighbors` function can
        be used to automatically compute and store all neighbor lists required by a
        given model.
    """

    def __init__(self, options: NeighborListOptions, length_unit: str):
        """
        :param options: :py:class:`metatensor.torch.atomistic.NeighborListOptions`
            defining the parameters of the neighbor list
        :param length_unit: unit of length used for the systems data

        Example
        -------

        >>> from vesin.torch.metatensor import NeighborList
        >>> from metatensor.torch.atomistic import System, NeighborListOptions
        >>> import torch
        >>> system = System(
        ...     positions=torch.eye(3).requires_grad_(True),
        ...     cell=4 * torch.eye(3).requires_grad_(True),
        ...     types=torch.tensor([8, 1, 1]),
        ...     pbc=torch.ones(3, dtype=bool),
        ... )
        >>> options = NeighborListOptions(cutoff=4.0, full_list=True, strict=False)
        >>> calculator = NeighborList(options, length_unit="Angstrom")
        >>> neighbors = calculator.compute(system)
        >>> neighbors
        TensorBlock
            samples (18): ['first_atom', 'second_atom', 'cell_shift_a', 'cell_shift_b', 'cell_shift_c']
            components (3): ['xyz']
            properties (1): ['distance']
            gradients: None
        <BLANKLINE>

        The returned TensorBlock can then be registered with the system

        >>> system.add_neighbor_list(options, neighbors)
        """  # noqa: E501

        if not torch.jit.is_scripting():
            if not _HAS_METATENSOR:
                raise ModuleNotFoundError(
                    "`vesin.metatensor` requires the `metatensor-torch` package"
                )
        self.options = options
        self.length_unit = length_unit
        self._nl = NeighborListTorch(
            cutoff=self.options.engine_cutoff(self.length_unit),
            full_list=self.options.full_list,
        )

        # cached Labels
        self._components = [Labels("xyz", torch.tensor([[0], [1], [2]]))]
        self._properties = Labels(["distance"], torch.tensor([[0]]))

    def compute(self, system: System) -> TensorBlock:
        """
        Compute the neighbor list for the given
        :py:class:`metatensor.torch.atomistic.System`.

        :param system: a :py:class:`metatensor.torch.atomistic.System` containing the
            data about a structure. If the positions or cell of this system require
            gradients, the neighbors list values computational graph will be set
            accordingly.

            The positions and cell need to be in the length unit defined for this
            :py:class`NeighborList` calculator.
        """

        # move to float64, as vesin only works in torch64
        points = system.positions.to(torch.float64)
        box = system.cell.to(torch.float64)
        if torch.all(system.pbc):
            periodic = True
        elif not torch.any(system.pbc):
            periodic = False
        else:
            raise NotImplementedError(
                "vesin currently does not support mixed periodic and non-periodic "
                "boundary conditions"
            )

        # computes neighbor list
        (P, S, D) = self._nl.compute(
            points=points, box=box, periodic=periodic, quantities="PSD", copy=True
        )

        # converts to a suitable TensorBlock format
        neighbors = TensorBlock(
            D.reshape(-1, 3, 1).to(system.positions.dtype),
            samples=Labels(
                names=[
                    "first_atom",
                    "second_atom",
                    "cell_shift_a",
                    "cell_shift_b",
                    "cell_shift_c",
                ],
                values=torch.hstack([P, S]),
            ),
            components=self._components,
            properties=self._properties,
        )

        return neighbors



from typing import List

import torch


class NeighborList:
    """A neighbor list calculator that can be used with TorchScript."""

    def __init__(self, cutoff: float, full_list: bool):
        """
        :param cutoff: spherical cutoff for this neighbor list
        :param full_list: should we return each pair twice (as ``i-j`` and ``j-i``) or
            only once
        """
        self._c = torch.classes.vesin._NeighborList(cutoff=cutoff, full_list=full_list)

    def compute(
        self,
        points: torch.Tensor,
        box: torch.Tensor,
        periodic: bool,
        quantities: str,
        copy: bool = True,
    ) -> List[torch.Tensor]:
        """
        Compute the neighbor list for the system defined by ``positions``, ``box``, and
        ``periodic``; returning the requested ``quantities``.

        ``quantities`` can contain any combination of the following values:

        - ``"i"`` to get the index of the first point in the pair
        - ``"j"`` to get the index of the second point in the pair
        - ``"P"`` to get the indexes of the two points in the pair simultaneously
        - ``"S"`` to get the periodic shift of the pair
        - ``"d"`` to get the distance between points in the pair
        - ``"D"`` to get the distance vector between points in the pair

        :param points: positions of all points in the system
        :param box: bounding box of the system
        :param periodic: should we use periodic boundary conditions?
        :param quantities: quantities to return, defaults to "ij"
        :param copy: should we copy the returned quantities, defaults to ``True``.
            Setting this to ``False`` might be a bit faster, but the returned tensors
            are view inside this class, and will be invalidated whenever this class is
            garbage collected or used to run a new calculation.

        :return: list of :py:class:`torch.Tensor` as indicated by ``quantities``.
        """

        return self._c.compute(
            points=points,
            box=box,
            periodic=periodic,
            quantities=quantities,
            copy=copy,
        )



from typing import List

import pytest
import torch

from vesin.torch import NeighborList


def test_errors():
    points = torch.tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 2.0]], dtype=torch.float64)
    box = torch.zeros((3, 3), dtype=torch.float64)

    calculator = NeighborList(cutoff=2.8, full_list=True)

    message = "only float64 dtype is supported in vesin"
    with pytest.raises(ValueError, match=message):
        calculator.compute(
            points.to(torch.float32),
            box.to(torch.float32),
            periodic=False,
            quantities="ij",
        )

    message = "expected `points` and `box` to have the same dtype, got Double and Float"
    with pytest.raises(ValueError, match=message):
        calculator.compute(
            points,
            box.to(torch.float32),
            periodic=False,
            quantities="ij",
        )

    message = "expected `points` and `box` to have the same device, got cpu and meta"
    with pytest.raises(ValueError, match=message):
        calculator.compute(
            points,
            box.to(device="meta"),
            periodic=False,
            quantities="ij",
        )

    message = "unexpected character in `quantities`: Q"
    with pytest.raises(ValueError, match=message):
        calculator.compute(
            points,
            box,
            periodic=False,
            quantities="ijQ",
        )

    message = "device meta is not supported in vesin"
    with pytest.raises(RuntimeError, match=message):
        calculator.compute(
            points.to(device="meta"),
            box.to(device="meta"),
            periodic=False,
            quantities="ij",
        )


@pytest.mark.parametrize("quantities", ["ijS", "D", "d", "ijSDd"])
def test_all_alone_no_neighbors(quantities):
    points = torch.tensor([[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]], dtype=torch.float64)
    box = torch.eye(3, dtype=torch.float64)

    calculator = NeighborList(cutoff=0.1, full_list=True)
    outputs = calculator.compute(points, box, True, quantities)

    if "ij" in quantities:
        assert list(outputs[quantities.index("i")].shape) == [0]
        assert list(outputs[quantities.index("j")].shape) == [0]

    if "S" in quantities:
        assert list(outputs[quantities.index("S")].shape) == [0, 3]

    if "D" in quantities:
        assert list(outputs[quantities.index("D")].shape) == [0, 3]
        assert not outputs[quantities.index("D")].requires_grad

    if "d" in quantities:
        assert list(outputs[quantities.index("d")].shape) == [0]
        assert not outputs[quantities.index("d")].requires_grad

    points.requires_grad_(True)
    box.requires_grad_(True)
    outputs = calculator.compute(points, box, True, quantities)

    if "ij" in quantities:
        assert list(outputs[quantities.index("i")].shape) == [0]
        assert list(outputs[quantities.index("j")].shape) == [0]

    if "S" in quantities:
        assert list(outputs[quantities.index("S")].shape) == [0, 3]

    if "D" in quantities:
        assert list(outputs[quantities.index("D")].shape) == [0, 3]
        assert outputs[quantities.index("D")].requires_grad

    if "d" in quantities:
        assert list(outputs[quantities.index("d")].shape) == [0]
        assert outputs[quantities.index("d")].requires_grad


class NeighborListWrap:
    def __init__(self, cutoff: float, full_list: bool):
        self._c = NeighborList(cutoff=cutoff, full_list=full_list)

    def compute(
        self,
        points: torch.Tensor,
        box: torch.Tensor,
        periodic: bool,
        quantities: str,
        copy: bool,
    ) -> List[torch.Tensor]:
        return self._c.compute(
            points=points,
            box=box,
            periodic=periodic,
            quantities=quantities,
            copy=copy,
        )


def test_script():
    class TestModule(torch.nn.Module):
        def forward(self, x: NeighborListWrap) -> NeighborListWrap:
            return x

    module = TestModule()
    module = torch.jit.script(module)



from typing import List, Optional, Union

import torch

from ._neighbors import NeighborList


try:
    from metatensor.torch.atomistic import (
        MetatensorAtomisticModel,
        ModelInterface,
        NeighborListOptions,
        System,
    )

    _HAS_METATENSOR = True
except ModuleNotFoundError:
    _HAS_METATENSOR = False

    class MetatensorAtomisticModel:
        pass

    class ModelInterface:
        pass

    class NeighborListOptions:
        pass

    class System:
        pass


def compute_requested_neighbors(
    systems: Union[List[System], System],
    system_length_unit: str,
    model: Union[MetatensorAtomisticModel, ModelInterface],
    model_length_unit: Optional[str] = None,
):
    """
    Compute all neighbors lists requested by the ``model`` through
    ``requested_neighbor_lists()`` member functions, and store them inside all the
    ``systems``.

    :param systems: Single system or list of systems for which we need to compute the
        neighbor lists that the model requires.
    :param system_length_unit: unit of length used by the data in ``systems``
    :param model: :py:class:`MetatensorAtomisticModel` or any ``torch.nn.Module``
        following the :py:class:`ModelInterface`
    :param model_length_unit: unit of length used by the model, optional. This is only
        required when giving a raw model instead of a
        :py:class:`MetatensorAtomisticModel`.
    """

    if isinstance(model, MetatensorAtomisticModel):
        if model_length_unit is not None:
            if model.capabilities().length_unit != model_length_unit:
                raise ValueError(
                    f"the given `model_length_unit` ({model_length_unit}) does not "
                    f"match the model capabilities ({model.capabilities().length_unit})"
                )

        all_options = model.requested_neighbor_lists()
    elif isinstance(model, torch.nn.Module):
        if model_length_unit is None:
            raise ValueError(
                "`model_length_unit` parameter is required when not "
                "using MetatensorAtomisticModel"
            )

        all_options = []
        _get_requested_neighbor_lists(
            model, model.__class__.__name__, all_options, model_length_unit
        )

    if not isinstance(systems, list):
        systems = [systems]

    for options in all_options:
        calculator = NeighborList(options, system_length_unit)
        for system in systems:
            neighbors = calculator.compute(system)
            system.add_neighbor_list(options, neighbors)


def _get_requested_neighbor_lists(
    module: torch.nn.Module,
    module_name: str,
    requested: List[NeighborListOptions],
    length_unit: str,
):
    """
    Recursively extract the requested neighbor lists from a non-exported metatensor
    atomistic model.
    """
    if hasattr(module, "requested_neighbor_lists"):
        for new_options in module.requested_neighbor_lists():
            new_options.add_requestor(module_name)

            already_requested = False
            for existing in requested:
                if existing == new_options:
                    already_requested = True
                    for requestor in new_options.requestors():
                        existing.add_requestor(requestor)

            if not already_requested:
                if new_options.length_unit not in ["", length_unit]:
                    raise ValueError(
                        f"NeighborsListOptions from {module_name} already have a "
                        f"length unit ('{new_options.length_unit}') which does not "
                        f"match the model length units ('{length_unit}')"
                    )

                new_options.length_unit = length_unit
                requested.append(new_options)

    for child_name, child in module.named_children():
        _get_requested_neighbor_lists(
            module=child,
            module_name=module_name + "." + child_name,
            requested=requested,
            length_unit=length_unit,
        )
